#ifdef _WIN32
#  define WINDOWS_LEAN_AND_MEAN
#  define NOMINMAX
#  include <windows.h>
#endif

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, GL
#include <GL/glew.h>
#include <GL/glut.h>

// includes
#include <cutil.h>
#include <cutil_gl_error.h>
#include <cuda_gl_interop.h>

/////////////////////////////
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
/////////////////////////////


////////////////////////////////////////////////////////////////////////////////
unsigned int window_width;
unsigned int window_height;

unsigned int mesh_width;
unsigned int mesh_height;


float * mesh;
float * dptr;

float anim = 0.0;

// network info
int sanity=0;
int portNumber;
int sockfd;
////////////////////////////////////////////////////////////////////////////////
// kernels
#include <dyn_simpleGL_kernel.cu>

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest( int argc, char** argv);

// Cuda functionality
void runCuda( );

// Schaber functionality
void waitForConnection(int port);
void initEnvironmentVariables(void);
void requestLoop(void);
void init(void);

#define GETNEWFRAME 1
#define TERMINATE 3

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv)
{
    waitForConnection(atoi(argv[1]));
    initEnvironmentVariables();

    init( );
    requestLoop( );
}

/**********Network Functions*****************/
void waitForConnection(int port) {
   int mySocket; 
   if((mySocket=socket(PF_INET,SOCK_STREAM,0))==-1) {
      perror("socket()");
      exit(1);
   }

   int yes=1;
   if(setsockopt(mySocket,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int))==-1) {
      perror("setsockopt()");
      exit(1);
   }

   struct sockaddr_in my_addr;
   my_addr.sin_family=AF_INET;
   my_addr.sin_port=htons(port);
   my_addr.sin_addr.s_addr=INADDR_ANY;
   memset(my_addr.sin_zero, '\0', sizeof my_addr.sin_zero);

   if(bind(mySocket,(struct sockaddr*)&my_addr, sizeof(struct sockaddr))==-1) {
      perror("bind()");
      exit(1);
   }

   if(listen(mySocket,5)==-1) {
      perror("listen()");
      exit(1);
   }

   struct sockaddr_in remote_addr;
   int sin_size=sizeof remote_addr;
   
   fprintf(stdout,"Waiting for connection on port %d...\n",port);
   if((sockfd=accept(mySocket, (struct sockaddr*)&remote_addr, (socklen_t*)&sin_size))==-1) {
      perror("accept()");
      exit(1);
   }

   fprintf(stdout, "Accepted connection from %s\n",inet_ntoa(remote_addr.sin_addr));
}

void initEnvironmentVariables(void) {
   int dim;
   recv(sockfd, &dim, 4, 0);
   printf("dim %d\n", dim);
   mesh_width = dim;
   mesh_height = dim;
   window_width = 2*dim;
   window_height = 2*dim;
   /*If client sends preferences overwrite defaults*/
}

void requestLoop() {

   char requestBuffer[18];
   int bytesRead,bytesSent;
   char request;

   fprintf(stdout,"Waiting for requests...\n");
   while(true) {
      bytesRead=recv(sockfd, (void*)requestBuffer,18,0);
      if(bytesRead==-1) {
         perror("recv()");
         exit(0);
      }
      if(bytesRead==0)
      {
	 printf("Read 0 bytes while requesting a new frame\n");
         exit(0);
      }

      request=requestBuffer[0];

      if(request==GETNEWFRAME) {
         fprintf(stderr,"New frame requested\n");
         runCuda( );
	 anim += 0.1;
         bytesSent=send(sockfd,(void*)mesh,mesh_width * mesh_height * 4 * sizeof(float),0);
         if(bytesSent==-1) {perror("send()"); exit(1);}
         fprintf(stderr,"Sent %d bytes\n",bytesSent);
      }
      else if(request==TERMINATE) {
         fprintf(stdout,"Recieved request to terminate. Exiting...\n");
         exit(0);
      }
      else {
         fprintf(stderr,"Recieved invalid command %d\n",request);
         exit(1);
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
//! Run the Cuda part of the computation
////////////////////////////////////////////////////////////////////////////////
void runCuda( )
{
    // execute the kernel
    dim3 block(8, 8, 1);
    dim3 grid(mesh_width / block.x, mesh_height / block.y, 1);
    kernel<<< grid, block>>>(dptr, mesh_width, mesh_height, anim);

    cudaMemcpy(mesh, dptr, mesh_width * mesh_height * 4 * sizeof(float), cudaMemcpyDeviceToHost);
}

void init( )
{
    int size = mesh_width * mesh_height * 4 * sizeof(float);
    mesh = (float *)malloc(mesh_width * mesh_height * 4 * sizeof(float));
    cudaMalloc((void **) &dptr, size);
}
